//  $Id: cmdopt.cpp,v 1.3 2003/04/28 05:49:15 t Exp $
//
///  MakLib Code Library
//
//  cmdopt.cpp
//
///  <overview>
///  <jpn>R}hCIvV</jpn>
///  <eng>Command-line option processing </eng>
///  </overview>
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "config_maklib.h"

#include <sstream>
#include <iomanip>
#include <climits>
#include <cstring>
#include "cmdopt.h"

using namespace std;

#if defined(HAVE_STRCASECMP)
inline int stricmp(string s1, string s2) { return strcasecmp(s1.c_str(), s2.c_str()); }
#elif defined(HAVE_STRICMP)
inline int stricmp(string s1, string s2) { return stricmp(s1.c_str(), s2.c_str()); }
#elif defined(HAVE_STRCMPI)
inline int stricmp(string s1, string s2) { return strcmpi(s1.c_str(), s2.c_str()); }
#else
inline int stricmp(string s1, string s2) { return strcmp(s1.c_str(), s2.c_str()); }
#endif

// #define DEBUG_PARSER

#ifdef DEBUG_PARSER
string disp(char c)
{
	if( c > ' ' ) return "'" + string(1,c) + "'";
	ostringstream oss;
	oss << "'\\x" << hex << setw(2) << setfill('0') << (int)c << "'";
	return oss.str();
}

string disp(bool b)
{
	return b ? "true" : "false";
}
#endif


namespace mak {

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt_base member function

void cmdopt_base::usage(ostream& os) const
{
	if( opt )
		os << "  -" << opt ;
	else
		os << "    ";

	if( name != "" )
		os << ", --" << name << usage_valstr() << endl << "     ";
	else
		os << usage_valstr() << "   ";

//	os << _usage << endl;
	istringstream iss( _usage );

	bool firstline = true;
	while( iss )
	{
		string s;
		getline(iss, s, '\n');
		if( iss && s != "" )
		{
			if( firstline )
				firstline = false;
			else
				os << "        ";
			os << s << endl;
		}
	}
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt<int> member function

bool cmdopt<int>::parse(istream& is, char delim)
{
	char c;
	val_t v;
	bool delim_kept = true;
	
	if( !is.good() || is.eof() || ! is.get(c) )
		throw exception_cmdopt("No value specified");
	else 
	{
		if( isspace(c) || c == delim )
		{
			delim_kept = false;
			if( c != delim )
			{
//				std::cout << "Ha-ha " << (int)c << endl;
				while( is.good() && !is.eof() && isspace(c) )
					if( !is.get(c) ) { c = '\0'; break; }
//				std::cout << "Huey " << (int)c << endl;
			}
			else
			{
//				std::cout << "Hello " << (int)c << endl;
				if( ! is.get(c) )
					c = '\0'; //// todo: ̂ւ̂II
//				std::cout << "Bye " << (int)c << endl;
			}
			if( isdigit(c) || c == '-' )
			{
				is.putback(c);
				is >> v;
				delim_kept = true;
			}
			else
				throw exception_cmdopt("Value required for option --" + longopt() + ": found '" + c + "'");
		}
		else if( c == '=' )
			is >> v;
		else
		{
			is.putback(c);
			if( isdigit(c) || c == '-' )
				is >> v;
			else
				throw exception_cmdopt("Value required for option --" + longopt() + ": found '" + c + "'");
		}
	}
	if( !is.fail() )
	{
		if( v < min )
		{
			ostringstream oss;
			oss << "The value of option --" << longopt();
			if( charopt() != '\0' )
				oss << " (-" + string(1,charopt()) + ")";
			oss << " is ";
				oss << "at least "<< min;

			throw exception_cmdopt(oss.str());
		}
		else if( v > max )
		{
			ostringstream oss;
			oss << "The value of option --" << longopt();
			if( charopt() != '\0' )
				oss << " (-" + string(1,charopt()) + ")";
			oss << " is ";
			oss << "at most "<< max;

			throw exception_cmdopt(oss.str());
		}
		set(v);
	}
	return delim_kept;
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt<string> member function

bool cmdopt<string>::parse(istream& is, char delim)
{
	char c;
	val_t v;
	bool retval;

	if( !is.good() )
		throw exception_cmdopt("Option processing failed due to bad condition of input stream");
	if( is.eof() || ! is.get(c) )
		throw exception_cmdopt("A string is required");
	else 
	{
		if( delim ? c == delim : isspace(c) )
		{
			if( delim )
			{
				if( ! getline(is, v, delim) )
					throw exception_cmdopt("A string is required");
				if( v.size() && v[0] == '-' )
					throw exception_cmdopt("A string is required (use --opt=value syntax for hyphen-leading strings)");
#ifdef DEBUG_PARSER
				cout << "DP string: get after delim, v=\"" << v << "\"" << endl;
#endif
				if( is.fail() )	// When no delim at last, clear error 
				{
					is.clear();
#ifdef DEBUG_PARSER
					cout << "DP string: bad cond" << endl;
#endif
				}
				retval = false;
			}
			else
			{
				is >> v;
#ifdef DEBUG_PARSER
				cout << "DP string: get after no delim, v=\"" << v << "\"" << endl;
#endif
				if( !is ) 
					throw exception_cmdopt("A string is required");
				if( v.size() && v[0] == '-' )
					throw exception_cmdopt("A string is required (use --opt=value syntax for hyphen-leading strings)");
				retval = true;
			}
		}
		else
		{
			bool quote = false;
			if( c != '=' )
				is.putback(c);
			if( ! delim )	// f~^ȂꍇA͂NH[e[V`FbN
			{
				if( is.get(c) ) 
				{	// Ȃ eof ̏ꍇ͂ȂɂȂ
					if( c == '\'' || c == '\"' )
					{
#ifdef DEBUG_PARSER
						cout << "DP string: quoted by = " << disp(c) << endl;
#endif
						quote = true;
						delim = c;
					}
					else
						is.putback(c);
				}
			}
			v = "";
			if( is.good() )
			{
				if( delim )
				{
					getline(is, v, delim);
#ifdef DEBUG_PARSER
					cout << "DP string: delim=" << disp(delim) << ", v=\"" << v << "\"" << endl;
#endif
					retval = quote;
				}
				else
				{
					if( isspace(c) )	// space after equal, e.g. "--str= " when no delim
					{
#ifdef DEBUG_PARSER
						cout << "DP string: no delim and trail space, v=\"\"" << endl;
#endif
						v = "";
					}
					else
					{
#ifdef DEBUG_PARSER
						cout << "DP string: no delim, v=\"" << v << "\"" << endl;
#endif
						is >> v;
					}

					retval = true;
				}
			}
			else
			{
#ifdef DEBUG_PARSER
				cout << "DP string: EOF after =, set null" << endl;
#endif
				if( is.fail() )
					is.clear();
				retval = false;
			}
		}
	}
	if( ! is.fail() )
	{
#ifdef DEBUG_PARSER
		cout << "DP string: result = \"" << v << "\"" << endl;
#endif
		set(v);
	}
#ifdef DEBUG_PARSER
	cout << "DP string: return " << disp(retval) << endl;
#endif
	return retval;
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt<bool> member function

bool cmdopt<bool>::parse(istream& is, char delim)
{
	char c;
	val_t v;
	if( !is.good() || ! is.get(c) )
	{
#ifdef DEBUG_PARSER
		cout << "DP bool: at eof" << endl;
#endif
		is.clear();
		v = true;
	}
	else 
	{
#ifdef DEBUG_PARSER
		cout << "DP bool: " << disp(c) << endl;
#endif
		if( isspace(c) )
		{
			v = true;
#ifdef DEBUG_PARSER
			cout << "DP bool: PUTBACK 1" << endl;
#endif
			is.putback(c);
		}
		else if( c == '+' || c == '1' )
			v = true;
		else if( c == '-' || c == '0' )
			v = false;
		else if( c == '=' )
		{
			string s;
			if( delim )
			{
				getline(is, s, delim);
				is.putback(delim);
			}
			else
				is >> s;

#ifdef DEBUG_PARSER
			cout << "DP bool: " << s << endl;
#endif

			if( stricmp(s, "true") == 0 || stricmp(s, "yes") == 0 || stricmp(s, "on") == 0 || s == "1" )
				v = true;
			else if( stricmp(s, "false") == 0 || stricmp(s, "no") == 0 || stricmp(s, "off") == 0 || s == "0" )
				v = false;
			else
				throw exception_cmdopt("Invalid value for boolean option");
		}
		else {
#ifdef DEBUG_PARSER
			cout << "DP bool: PUTBACK 2" << endl;
#endif
			is.putback(c);
		}
	}
	if( !is.fail() )
		set(v);
#ifdef DEBUG_PARSER
	cout << "DP bool: return true" << endl;
#endif
	return true;
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt<double> member function

bool cmdopt<double>::parse(istream& is, char delim)
{
	char c;
	val_t v;
	bool delim_kept = true;
	
	if( is.bad() || is.eof() || ! is.get(c) )
		throw exception_cmdopt("No value specified");
	else 
	{
		if( isspace(c) || c == delim )
		{
			delim_kept = false;
			if( c != delim )
				while( isspace(c) )
				{
					if( ! is.get(c) ) 
						{ c = '\0'; break; }
				}
			else
				if( ! is.get(c) )
					c = '\0'; //// todo: ̂ւ̂II
			if( isdigit(c) || c == '-' || c == '.')
			{
				is.putback(c);
				is >> v;
				delim_kept = true;
			}
			else
				throw exception_cmdopt("Value required for option --" + longopt());
		}
		else if( c == '=' )
			is >> v;
		else
		{
			is.putback(c);
			if( isdigit(c) || c == '-' || c == '.' )
				is >> v;
			else
			throw exception_cmdopt("number is required for option --" + longopt());
		}
	}
	if( !is.fail() )
	{
		if( v < min )
		{
			ostringstream oss;
			oss << "The value of option --" << longopt();
			if( charopt() != '\0' )
				oss << " (-" + string(1,charopt()) + ")";
			oss << " is ";
				oss << "at least "<< min;

			throw exception_cmdopt(oss.str());
		}
		else if( v > max )
		{
			ostringstream oss;
			oss << "The value of option --" << longopt();
			if( charopt() != '\0' )
				oss << " (-" + string(1,charopt()) + ")";
			oss << " is ";
			oss << "at most "<< max;

			throw exception_cmdopt(oss.str());
		}
		set(v);
//		cout << "mogemogemoge " << v << endl;
	}
	return delim_kept;
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt_help member function

bool cmdopt_help::parse(istream& /*is*/, char /*delim*/)
{
	parser->usage(cout);

	exit(1);
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt_version member function

bool cmdopt_version::parse(istream& /*is*/, char /*delim*/)
{
	cout << version << endl;

	exit(1);
}

///////////////////////////////////////////////////////////////////////////
//
//  cmdopt_parser member function

void cmdopt_parser::usage(ostream& os) const
{
	for( vec_t::const_iterator i = vec.begin(); i != vec.end(); i++ )
	{
		if( i->second )
			i->first->usage(os);
	}
}

int cmdopt_parser::parse(int argc, char **argv) throw( exception_cmdopt )
{
	stringstream ss;
	for( int i=1; i<argc; i++ )
	{
		ss << argv[i] << '\n';
	}
	return parse(ss, '\n');
}

int cmdopt_parser::parse(istream &is, char delim) throw( exception_cmdopt )
{
	ostringstream errstr;
	bool error = false;
	
#ifdef DEBUG_PARSER
	cout << "DP: start, delim = "<< disp(delim) << endl;
#endif
	while( is && ! is.eof() )
	{
//		cout << "hoge0" << endl;
		char c;
		if( ! is.get(c) )
			break;
#ifdef DEBUG_PARSER
		cout << "DP: 1st char taken: " << disp(c) << endl;
#endif
		if( c == '-' )
		{
			if( ! is.get(c) )
			{
#ifdef DEBUG_PARSER
				cout << "DP: Stream finished. store '-' to the argv" << endl;
#endif
				_argv.push_back("-");
				break;
			}
#ifdef DEBUG_PARSER
			cout << "DP: 2nd char taken: " << disp(c) << endl;
#endif
			if( c == '-' )
			{
				string cmd = "";
				while( !is.eof() && is.good() )
				{
					if( ! is.get(c) )
						break;
					if( (!delim ? isspace(c) : c==delim) || c == '=' )
						break;
					cmd += c;
				}
				is.putback(c);
#ifdef DEBUG_PARSER
				cout << "DP: -- option: [" << cmd << "], putbacked " << disp(c) << endl;
#endif
				strmap_t::iterator i = strmap.find(cmd);
				if( i == strmap.end() )
				{
#ifdef DEBUG_PARSER
					cout << "DP: --[" << cmd << "] not found" << endl;
#endif
					error = true; errstr << "Undefined option '--" + cmd + "'. ";
					if( delim )
						is.ignore(INT_MAX, delim);
					else
					{
						string dummy;
						is >> dummy;
					}
				}
				else
				{
#ifdef DEBUG_PARSER
					cout << "DP: --[" << cmd << "] found" << endl;
#endif
					bool ret;
					try {
						ret = i->second->parse(is, delim);
#ifdef DEBUG_PARSER
						cout << "DP: --[" << cmd << "] success, return=" << disp(ret) << endl;
#endif
					} 
					catch( exception &e ) {
#ifdef DEBUG_PARSER
						cout << "DP: --[" << cmd << "] failed, reason=" << e.what() << endl;
#endif
						error = true; errstr << "Error in option '--" + cmd + "': " << e.what();
						ret = false;
					}
					if( ret && !is.eof() )
					{
#ifdef DEBUG_PARSER
						c = '\0';
#endif
						if( is.get(c) && !(delim ? c == delim : isspace(c)) )	// Check tail of the option value
						{
#ifdef DEBUG_PARSER
							cout << "DP: --[" << cmd << "] trail error " << disp(c) << endl;
#endif
							error = true; errstr << "Bad value in option '--" + cmd + "': trailed char is '" << c << "'. ";
						} 
#ifdef DEBUG_PARSER
						else 
							cout << "DP: --[" << cmd << "] trail no-error " << disp(c) << endl;
#endif
					}
					else if( is.fail() )
					{
#ifdef DEBUG_PARSER
						cout << "DP: --[" << cmd << "] stream bad" << endl;
#endif
						error = true; errstr << "Error in option '--" + cmd + "': Invalid value. ";
						break;
					}
#ifdef DEBUG_PARSER
					else
						cout << "DP: --[" << cmd << "] good" << endl;
#endif
				}
			}
			else
			{
				if( isspace(c) || c == EOF ) {
#ifdef DEBUG_PARSER
					cout << "DP: single -, pushed to argv" << endl;
#endif
					_argv.push_back("-");
				}
				else
				{
					do {
						charmap_t::iterator i = charmap.find(c);
						if( i == charmap.end() )
						{
#ifdef DEBUG_PARSER
							cout << "DP: -" << disp(c) << " not found" << endl;
#endif
							error = true; errstr << "Undefined option '-" + string(1,c) + "'. ";
						}
						else
						{
#ifdef DEBUG_PARSER
							cout << "DP: -" << disp(c) << " found" << endl;
#endif
							bool ret;
							try {
								ret = i->second->parse(is, delim);
#ifdef DEBUG_PARSER
							cout << "DP: -" << disp(c) << " success, ret=" << disp(ret) << endl;
#endif
							} 
							catch( exception &e ) {
#ifdef DEBUG_PARSER
								cout << "DP: -" << disp(c) << " failed, reason=" << e.what() << endl;
#endif
								error = true; errstr << "Error in option '-" + string(1,c) + "': " << e.what();
								ret = false;
							}
							if( is.fail() )
							{
#ifdef DEBUG_PARSER
								cout << "DP: -" << disp(c) << " stream bad"<< endl;
#endif
								error = true; errstr << "Error in option '-" + string(1,c) + "': Invalid value. ";
								break;
							}
							if( ret == false ) {
#ifdef DEBUG_PARSER
								cout << "DP: -" << disp(c) << " delim over, exit inner loop"<< endl;
#endif
								break;	// Over delimiter; exit loop
							}
						}
						is.get(c);
#ifdef DEBUG_PARSER
						cout << "DP: next char: " << disp(c) << endl;
#endif
					} while( ! is.eof() && !(delim ? c == delim : isspace(c)) );
				}
			}
		}
		else if( ! isspace(c) /*&& c != EOF*/)
		{
			is.putback(c);

			string val;
			if( delim )
				getline(is, val, delim);
			else
				is >> val;
			if( val.size() >= 1 && val[0] )
			{
#ifdef DEBUG_PARSER
				cout << "DP: non-option: " << val << endl;
#endif
				_argv.push_back(val);
			}
		}
#ifdef DEBUG_PARSER
		cout << "DP: end of one argv "<< endl;
#endif
	}
#ifdef DEBUG_PARSER
	cout << "DP: finish"<< endl;
#endif
	if( error )
	{
//		cout << "hoge99" << endl;
		throw exception_cmdopt(errstr.str());
	}
//	cout << "hoge999" << endl;
	return _argv.size();
}

string cmdopt_parser::makeString() const
{
	string ret;
	string delim = "";
	for( vec_t::const_iterator i = vec.begin(); i != vec.end(); i++ )
	{
		string s = i->first->makeString();
		if( s != "" )
		{
			ret = ret + delim + s;
			delim = " ";
		}
	}
	return ret;
}

/*
// test code
int main(int argc, char **argv)
{
	cmdopt_parser p;

	cmdopt<int> cnt('c', "count", 0, 
		"Counts something.");
	cmdopt<string> set('s', "set", "", 
		"Sets something.");
	cmdopt<bool> bit('b', "bit", false, 
		"Flags something.");

	p.add( ref(cnt) );
	p.add( ref(set) );
	p.add( ref(bit) );
	p.add( New<cmdopt_help>('h', "help", p,
		"Show this help.") );

	try {
		p.parse(argc, argv);
	}
	catch(exception &e)
	{
		cerr << "Errors in options: " << endl << e.what();
		return 1;
	}
//	p.parse(argc, argv);


	cerr << "cnt value: " << count.get() << endl;
	cerr << "set value: " << set.get() << endl;
	cerr << "bit value: " << bit.get() << endl;

	vector<string> s = p.argv();
	cerr << "argv: " << endl;
	for( vector<string>::iterator i = s.begin(); i != s.end(); i++ )
	{
		cerr << "'" << *i << "'" << endl;
	}
	return 0;
}
*/

} // namespace mak
