//    $Id: memchk.cpp,v 1.3 2003/04/28 05:49:15 t Exp $
//
//    memchk.cpp: Memory-leak and overrun checking routines
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

static const char rcsid[] = "$Id: memchk.cpp,v 1.3 2003/04/28 05:49:15 t Exp $";

#include "config_maklib.h"

#include <cstdlib>
#include <iostream>
#ifdef HAVE_LIMITS_H
#  include <limits.h>
#endif
#include <exception>

#define SIZE_INVALID ((size_t)-1)

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#else
#ifdef HAVE_IO_H
#  include <io.h>
#else
extern "C" size_t write(int, const void *, size_t);
#endif
#endif

#include <memory>

#include "memchk.h"

size_t init_newcounter = 0;
size_t newcounter = 0;
size_t delcounter = 0;

void SanityCheck() throw();
void DeleteOldGarbage() throw();

void abesh() throw()
{
	std::cerr << "abesh " << newcounter << " " << delcounter << std::endl;
	SanityCheck();
}

#ifdef HAS_INT64
#define MAGICNUM_S 0xaf02c47918d326a9L
#define MAGICNUM_V 0xb31ae06d75326b97L
#define MAGICNUM_D 0xc3107f6d23e591feL
#else
#define MAGICNUM_S 0xaf02c479
#define MAGICNUM_V 0xb31ae06d
#define MAGICNUM_D 0xdddddddd
#endif

//#define TRAP_NEWCOUNTER 47
//#define TRAP_DELCOUNTER 15687


#ifndef DEBUG_MEM2
#define DISP_ALLOC(i) 
#define DISP_DELETE(i)
#else
#define DISP_ALLOC(i) disp("ALLOC  ", i)
#define DISP_DELETE(i) disp("DELETE ", i)


static const int ppp[] = { 100000000, 10000000, 1000000, 100000, 10000, 1000, 100, 10, 1 };

static void dispint(int i) throw()
{
	for( int n=0; n<9; n++ )
	{
		char c = i / ppp[n] + '0';
		write(2, &c, 1);
		i -= (c-'0') * ppp[n];
	}
}

static void disp(char *p, int i) throw()
{
	write(2, p, 7);
	
	dispint(i);
/*	write(2, " ", 1);
	dispint(newcounter);
	write(2, " ", 1);
	dispint(delcounter); */
	write(2, "\n", 1);
}
#endif


struct nheader
{
	size_t  newc;
	nheader *next;
	nheader *prev;
	size_t  size;
};

nheader *mem_root = NULL;
nheader *delete_root = NULL;

static void init() 
{
	std::cerr << "Initialize " << __FILE__ << std::endl;
	init_newcounter = newcounter;
}

static void term() 
{ 
	std::cerr << "newcounter: " << newcounter << std::endl;
	std::cerr << "delcounter: " << delcounter << std::endl;
	SanityCheck();	

	nheader *nh = mem_root;
	int cnt = 0;
	if( nh != NULL )
		do {
			if( nh->newc >= init_newcounter )
			{
				if( ++cnt > 20 )
				{
					std::cerr << "more leaking block found." << std::endl;
					return;
				}
				std::cerr << "leaking block newc = " << nh->newc << std::endl;
			}
			nh = nh->next;
		} while( nh != mem_root );
}

class init_term
{
public:
	init_term() { init(); }
	~init_term() { term(); }
};

init_term init_term_obj;


void * operator new(size_t isize) throw(std::bad_alloc)
{
	++newcounter;
	if( (newcounter & (countcheck-1)) == 0 )
		SanityCheck();
#ifdef TRAP_NEWCOUNTER
	if( newcounter == TRAP_NEWCOUNTER )
		abesh();
#endif
	
	size_t size = (isize + 7) & ~7;
	if( size == 0 )
		size = 1;
	
	void *p = malloc(size + sizeof(size_t) + sizeof(nheader));
	
	if( p == NULL )
	{
		std::cerr << "virtual memory exceeded" << std::endl;
		throw std::bad_alloc();
	}

	DISP_ALLOC(newcounter);
	
	if( mem_root == NULL )
	{
		((nheader *)p)->prev = (nheader *)p;
		((nheader *)p)->next = (nheader *)p;
		mem_root = (nheader *)p;
	}
	else
	{
		((nheader *)p)->prev = mem_root->prev;
		((nheader *)p)->next = mem_root;
		mem_root->prev->next = (nheader *)p;
		mem_root->prev = (nheader *)p;
	}
	((nheader *)p)->newc = newcounter;
	((nheader *)p)->size = size;
	((size_t *)((char *)p+size+sizeof(nheader)))[0] = MAGICNUM_S;
	
	return (char *)p + sizeof(nheader);
}

void * operator new[](size_t isize)  throw(std::bad_alloc)
{
	++newcounter;
	if( (newcounter & (countcheck-1)) == 0 )
		SanityCheck();

	size_t size = (isize + 7) & ~7;
	if( size == 0 )
		size = 1;
	
	void *p = malloc(size + sizeof(size_t) + sizeof(nheader));
	
	if( p == NULL )
	{
		std::cerr << "virtual memory exceeded" << std::endl;
		throw std::bad_alloc();
	}

	DISP_ALLOC(newcounter);

	if( mem_root == NULL )
	{
		((nheader *)p)->prev = (nheader *)p;
		((nheader *)p)->next = (nheader *)p;
		mem_root = (nheader *)p;
	}
	else
	{
		((nheader *)p)->prev = mem_root->prev;
		((nheader *)p)->next = mem_root;
		mem_root->prev->next = (nheader *)p;
		mem_root->prev = (nheader *)p;
	}
	((nheader *)p)->newc = newcounter;
	((nheader *)p)->size = size;
	((size_t *)((char *)p+size+sizeof(nheader)))[0] = MAGICNUM_V;
	
	return (char *)p + sizeof(nheader);
}

void operator delete(void *p)  throw()
{
	if( p != NULL )
	{
		delcounter++;
		if( (delcounter & (countcheck-1)) == 0 )
		{
			SanityCheck();
			DeleteOldGarbage();
		}

		nheader *nh = (nheader *)((char *)p - sizeof(nheader));
		size_t newc = nh->newc;
		if( newc > newcounter )
		{
			std::cerr << "Internal error: illegal delete, newc=" << newc << ", delc=" << delcounter << std::endl;
			return;
		}
		size_t size = nh->size;
		if( nh->next == NULL || nh->prev == NULL )
		{
			std::cerr << "Internal error: double delete, newc=" << newc << ", first delc=" << size << ", second delc=" << delcounter << std::endl;
			return;
		}
		size_t magicnum = ((size_t *)((char *)p + size))[0];

		if( magicnum != MAGICNUM_S )
		{
			if( magicnum == MAGICNUM_V )
				std::cerr << "Internal warning: delete for new[], newc=" << newc << ", delc=" << delcounter << std::endl;
			else
			{
				std::cerr << "Internal error: delete for unknown (memory overrun?), newc=" << newc << ", delc=" << delcounter << std::endl;
				return;
			}
		}
		if( nh->next->prev != nh || nh->prev->next != nh )
		{
			std::cerr << "Internal error: link broken, newc=" << newc << ", delc=" << delcounter << std::endl;
		}
		
		nh->size = delcounter;
		
		nh->next->prev = nh->prev;
		nh->prev->next = nh->next;

		if( nh->next == nh )
			mem_root = NULL;

		nh->prev = NULL;
		nh->next = delete_root;
		delete_root = nh;
		for( uint i=0; i<size / sizeof(size_t); i++ )
			((size_t *)p)[i] = MAGICNUM_D;
//		free( (char *)p - sizeof(size_t)*2 );
		DISP_DELETE(newc);
	}

	return;
}

void operator delete[](void *p) throw()
{
	if( p != NULL )
	{
		delcounter++;
		if( (delcounter & (countcheck-1)) == 0 )
		{
			SanityCheck();
			DeleteOldGarbage();
		}

		nheader *nh = (nheader *)((char *)p - sizeof(nheader));
		size_t newc = nh->newc;
		if( newc > newcounter )
		{
			std::cerr << "Internal error: illegal delete[], newc=" << newc << ", delc=" << delcounter << std::endl;
			return;
		}
		size_t size = nh->size;
		if( nh->next == NULL || nh->prev == NULL )
		{
			std::cerr << "Internal error: double delete[], newc=" << newc << ", first delc=" << size << ", second delc=" << delcounter << std::endl;
			return;
		}
		size_t magicnum = ((size_t *)((char *)p + size))[0];

		if( magicnum != MAGICNUM_V )
		{
			if( magicnum == MAGICNUM_S )
				std::cerr << "Internal warning: delete[] for new, newc=" << newc << ", delc=" << delcounter << std::endl;
			else
			{
				std::cerr << "Internal error: delete[] for unknown (memory overrun?), newc=" << newc << ", delc=" << delcounter << " (ptr: " << std::hex << (unsigned long)p << std::dec << std::endl;
				return;
			}
		}
		if( nh->next->prev != nh || nh->prev->next != nh )
		{
			std::cerr << "Internal error: link broken, newc=" << newc << ", delc=" << delcounter << std::endl;
		}
		
		nh->size = delcounter;
		
		nh->next->prev = nh->prev;
		nh->prev->next = nh->next;

		if( mem_root == nh )
		{
			if( nh->next == nh )
				mem_root = NULL;
			else
				mem_root = nh->next;
		}

		nh->prev = NULL;
		nh->next = delete_root;
		delete_root = nh;
		for( uint i=0; i<size / sizeof(size_t); i++ )
			((size_t *)p)[i] = MAGICNUM_D;
//		free( (char *)p - sizeof(size_t)*2 );
		DISP_DELETE(newc);
	}

	return;
}

void SanityCheck() throw()
{
//#if defined( DEBUG ) || defined( DEBUG_MEM2 )
	write(2, "SanityCheck", 11);
//#endif
	uint ncnt = 0;
	nheader *nh = mem_root;
	
	if( nh != NULL )
		do {
			if( ncnt++ > newcounter )
			{
				std::cerr << "Memory Insane ... counter over  newc:" << nh->newc << std::endl;
				return;
			}
			if( nh->newc > newcounter )
			{
				std::cerr << "Memory Insane ... bad newc" << std::endl;
				return;
			}
			if( nh->next == NULL )
			{
				std::cerr << "Memory Insane ... deleted link  newc:" << nh->newc << "  links:" << ncnt << std::endl;
				return;
			}
			if( nh->next->prev != nh )
			{
				std::cerr << "Memory Insane ... bad link  newc:" << nh->newc << "  links:" << ncnt << std::endl;
				return;
			}
			if( *(size_t *)((char *)nh+sizeof(nheader)+nh->size) != MAGICNUM_S && 
				*(size_t *)((char *)nh+sizeof(nheader)+nh->size) != MAGICNUM_V )
			{
				std::cerr << "MEMORY Insane ... magic number not match  newc:" << nh->newc << std::endl;
				return;
			}
			nh = nh->next;
		} while( nh != mem_root );
	write(2, "\010\010\010\010\010\010\010\010\010\010\010           \010\010\010\010\010\010\010\010\010\010\010", 33);
}

void DeleteOldGarbage() throw()
{
	write(2, "DeleteOldGB", 11);
	static nheader *pre_delete_root = NULL; 
	static nheader *prepre_delete_root = NULL; 

	while( prepre_delete_root != NULL )
	{
		nheader *p = prepre_delete_root->next;
		free( prepre_delete_root );
		prepre_delete_root = p;
	}

	prepre_delete_root = pre_delete_root;
	pre_delete_root = delete_root;
	delete_root = NULL;

	write(2, "\010\010\010\010\010\010\010\010\010\010\010           \010\010\010\010\010\010\010\010\010\010\010", 33);
}

