// $Id: test.cpp,v 1.4 2003/04/28 05:49:15 t Exp $
//
//    test.cpp: unit test routines
//
//    Copyright (c) 2003 Makino Takaki, t-libmak@snowelm.com
//
//    This library is free software; you can redistribute it and/or
//    modify it under the terms of the GNU Lesser General Public
//    License as published by the Free Software Foundation; either
//    version 2.1 of the License, or (at your option) any later version.
//
//    This library is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//    Lesser General Public License for more details.
//
//    You should have received a copy of the GNU Lesser General Public
//    License along with this library; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "test.h"

namespace mak {

using namespace std;

/////////////////////////////////////////////////////////////////////////
//  AssertFailedException

AssertFailedException::AssertFailedException(const char *teststr, const char *file, int line)
{ 
	ostringstream oss;
	oss << "Assertion Failed: " << file << ":" << line << ": " << teststr;
	char *ss = new char[oss.str().size()+1];
	strcpy( ss, oss.str().c_str() );
	s = ss;
}

AssertFailedException::~AssertFailedException() throw() { delete[] s; }

const char *AssertFailedException::what() { return s; }

void assertFail(const char *teststr, const char *file, int line) throw(AssertFailedException)
{
	throw AssertFailedException(teststr, file, line);
}


/////////////////////////////////////////////////////////////////////////
//  TestSuite

TestSuite *TestSuite::root = 0;


/////////////////////////////////////////////////////////////////////////
//  TestBase

void TestBase::_testEqual( string a, string b, const char *astr, const char *bstr, const char *file, int line )
{
	std::ostringstream st;
	st << file << ":" << line << ": testEqual " << astr << " (" << a << ") == " << bstr << " (" << b << ")";
	if( a == b ) {
		success(st.str());
	}
	else {
		failure(st.str());
	}
}

void TestBase::_testFailThrow( const char *astr, const char *bstr, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testThrow " << astr << " should throw " << bstr << ", but none was thrown";
	failure(st.str());
}

void TestBase::_testFailThrow( const char *astr, const char *bstr, std::exception &e, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testThrow " << astr << " should throw " << bstr << ", but another exception was thrown: " << e.what();
	failure(st.str());
}

void TestBase::_testFailThrowX( const char *astr, const char *bstr, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testThrow " << astr << " should throw " << bstr << ", but something different was thrown";
	failure(st.str());
}


void TestBase::_testSuccessThrow( const char *astr, const char *bstr, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testThrow " << astr << " throws " << bstr;
	success(st.str());
}

void TestBase::_testFailNoThrow( const char *astr, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testNoThrow(" << astr << ") should throw nothing";
	failure(st.str());
}

void TestBase::_testFailNoThrow( const char *astr, std::exception &e, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testNoThrow(" << astr << ") should throw nothing, but an exception was thrown: " << e.what();
	failure(st.str());
}

void TestBase::_testSuccessNoThrow( const char *astr, const char *file, int line )
{
	ostringstream st;
	st << file << ":" << line << ": testNoThrow(" << astr << ") throws no exception";
	success(st.str());
}

TestBase::TestBase(TestSuite *is /*= NULL*/) 
	: suite(NULL) 
{ 
	if( is != NULL ) 
		is->Add(this); 
}

TestBase::~TestBase() 
{ 
	if( suite != NULL ) 
		suite->Remove(this); 
}

void TestBase::success(std::string desc) throw()
{
	if( suite ) 
		suite->success(name() + (": " + desc));
	else
		std::cout << "Success: " << (name() + (": " + desc)) << endl;
}

void TestBase::failure(std::string desc) throw()
{
	if( suite ) 
		suite->failure(name() + (": " + desc));
	else
		std::cout << "Failure: " << (name() + (": " + desc)) << endl;
}

void TestBase::failureExpect(bool expect) throw() 
{
	if( suite ) 
		suite->failureExpect(expect);
}

/////////////////////////////////////////////////////////////////////
/// Test stuff

void Test::runTest() 
{
	try {
		setUp();
	}
	catch(exception &e)
	{
		failureExpect(false);
		failure("setUp() threw uncaught exception: " + std::string(e.what()));
		return;
	}
	catch(...)
	{
		failureExpect(false);
		failure("setUp() threw uncaught exception");
		return;
	}
	try {
		test();
	}
	catch(exception &e)
	{
		failureExpect(false);
		failure("test() threw uncaught exception: " + std::string(e.what()));
		return;
	}
	catch(...)
	{
		failureExpect(false);
		failure("test() threw uncaught exception");
		return;
	}
	try {
		tearDown();
	}
	catch(exception &e)
	{
		failureExpect(false);
		failure("tearDown() threw uncaught exception: " + std::string(e.what()));
		return;
	}
	catch(...)
	{
		failureExpect(false);
		failure("tearDown() threw uncaught exception");
		return;
	}
}


/////////////////////////////////////////////////////////////////////
/// TestBundle stuff

void TestBundle::runTest() 
{
	for( int i=1; i<=testCount(); i++ )
		runTest(i);
}

void TestBundle::runTest(int i) 
{
	try {
		setUp();
	}
	catch(exception &e) {
		failureExpect(false);
		failure("setUp() threw uncaught exception: " + std::string(e.what()));
		return;
	}
	catch(...) {
		failureExpect(false);
		failure("setUp() threw uncaught exception");
		return;
	}

	try {
		test(i);
	}
	catch(exception &e) {
		failureExpect(false);
		failure("test() threw uncaught exception: " + std::string(e.what()));
	}
	catch(...) {
		failureExpect(false);
		failure("test() threw uncaught exception");
	}

	try {
		tearDown();
	}
	catch(exception &e) {
		failureExpect(false);
		failure("tearDown() threw uncaught exception: " + std::string(e.what()));
		return;
	}
	catch(...) {
		failureExpect(false);
		failure("tearDown() threw uncaught exception");
		return;
	}
}


/////////////////////////////////////////////////////////////////////
/// TestSuite stuff

TestSuite::TestSuite(std::string iname) 
	: TestBase(), name(iname), os(NULL), succcnt(0), failcnt(0), unexcnt(0), verbose(1), failureExpected(false)
{ 
		next = root; 
		root = this; 
}

TestSuite::TestSuite(std::string iname, TestSuite *is) 
	: TestBase(is), name(iname), os(NULL), succcnt(0), failcnt(0), unexcnt(0), verbose(1), failureExpected(false)
{ 
	if( is == NULL ) { 
		next = root; 
		root = this; 
	} 
	else 
		next = NULL; 
}

TestSuite::~TestSuite() 
{ 
	if( suite == NULL )
	{
		if( root == this )
			root = next;
		else {
			TestSuite *p = root;
			while( p )
			{
				if( p->next == this )
				{
					p->next = next;
					break;
				}
				p = p->next;
			}
		}
	}
	for( tests_type::iterator i = tests.begin(); i != tests.end(); i++ )
	{
		(*i)->setSuite(NULL);
	}
}

void TestSuite::setUp()
{
	succcnt = 0;
	failcnt = 0;
	unexcnt = 0;
	failureExpected = false;
}

void TestSuite::runTest()
{
	for( tests_type::iterator i = tests.begin(); i != tests.end(); i++ )
	{
		if( verbose >= 3 )
			std::cout << "Start test: " << std::flush << (*i)->name() << std::endl;
		(*i)->runTest();
		if( verbose >= 3 )
			std::cout << "End   test: " << std::flush << (*i)->name() << std::endl;
	}
}

void TestSuite::tearDown()
{
	if( ! suite && verbose >= 1 )
	{
		std::cout << name << ": " << succcnt << " success(es), " << failcnt << " failure(s) (" << unexcnt << " unexpected failures)." << endl;
	}
}


void TestSuite::Add(TestBase *t)
{
	if( t->suite ) (t->suite)->Remove(t);
	tests.insert(t);	t->setSuite(this);
}

void TestSuite::Remove(TestBase *t)
{
	tests.erase(t);	t->setSuite(NULL);
}

void TestSuite::success(std::string desc) throw()
{
	succcnt++;
	if( suite ) 
		suite->success(name + ": " + desc);
	else
		if( verbose >= 2 )
			std::cout << "Success: " << (name + ": " + desc) << endl;
	failureExpected = false;
}

void TestSuite::failure(std::string desc) throw()
{
	failcnt++;
	if( ! failureExpected )
		unexcnt++;
	if( suite ) 
		suite->failure(name + ": " + desc);
	else
		if( verbose >= 1 )
			std::cout << "Failure: " << (name + ": " + desc) << endl;
	failureExpected = false;
}

void TestSuite::failureExpect(bool expect) throw()
{
	failureExpected = expect;
	if( suite ) 
		suite->failureExpect(expect);
}


void TestSuite::setStreamAll(ostream &o) 
{ 
	TestSuite *p = root; while(p) { p->setStream(o); p = p->next; } 
}
void TestSuite::setVerboseAll(int v) 
{ 
	TestSuite *p = root; while(p) { p->setVerbose(v); p = p->next; } 
}
int TestSuite::testAll() 
{ 
	int Success = 0;
	int Failure = 0;
	int unexpectedFailure = 0;
	TestSuite *p = root; 
	while(p) 
	{
		p->runTest(); 
		Success += p->getSuccessCount();
		Failure += p->getFailureCount();
		unexpectedFailure += p->getUnexpectedFailureCount();
		p = p->next; 
	} 
	std::cout << "Total: " << Success << " success(es), " << Failure << " failure(s) (" << unexpectedFailure << " unexpected failures)." << endl;
	if( Success+Failure == 0 )
	{
		std::cout << "Error: No test performed at all!" << std::endl;
	}
	return unexpectedFailure > 0 ? 1 : (Success+Failure == 0) ? 1 : 0;
}
int TestSuite::getSuccessCountAll() 
{ 
	int r = 0; const TestSuite *p = root; while(p) { r += p->getSuccessCount(); p = p->next; } return r; 
}
int TestSuite::getFailureCountAll() 
{ 
	int r = 0; const TestSuite *p = root; while(p) { r += p->getFailureCount(); p = p->next; } return r; 
}

}
